# -*- coding: utf-8 -*-

# Standard Library
import json
import datetime
import zlib
import base64
import socket
from dateutil.relativedelta import relativedelta
import re

# Vendor Library
from celery import group
from celery.exceptions import MaxRetriesExceededError
from django.template import loader
from django.db.models import Q
import suds
from suds.client import Client
from suds.sax.element import Element

# Project Library
from kinder.core.async.tasks import celery_app
from kinder.core.dict.models import (
    HNE, UnitKind, GroupTypeEnumerate, WorkType)
from kinder.core.group.models import Group, GroupStatusEnum
from kinder.core.emie_unit.models import EmieUnitModel
from kinder.core.unit.models import (
    Unit,
    FilialData,
    FilialTypeStatus)
from kinder.core.unit.models import BuildingTypes
from kinder.core.unit.models import LocalityEnumerate
from kinder.core.declaration.models import Declaration
from kinder.core.declaration_status.enum import DSS
from kinder.core.emie_unit.models import DopService
from gisdo.utils import UnitHelper
from gisdo.utils import get_file_url
from gisdo.utils import DateMixin
from gisdo.utils import AverageWaitingTimeWrapperEncoder
from gisdo.models import (
    ReportFormRow, ReportForm, ScheduleSettings, InstallationType)
from gisdo import settings as gs_settings
from gisdo.exceptions import (
    SendReportException, DocumentCreateException,
    XMLRenderException, SOAPClientCreateException,
    NoDataInMo, ConnectingAtSendingReportException)
from gisdo.logger import Logger
from gisdo.index.enrolled import Index
from .collector import ReportCacheMixin, DataCollector
from gisdo import constants as const
from .utils import UnitHelper
from .xml_helpers import get_dou_type
from .xml_helpers import get_dou_status
from .xml_helpers import get_dou_structure_type
from .xml_helpers import get_groups_num
from .xml_helpers import get_filial_num
from .xml_helpers import get_unit_fias_house_guid
from .xml_helpers import get_additional_info_by_dou
from .xml_helpers import get_filial_info


@celery_app.task(queue=gs_settings.CELERY_WORKER_TASK_QUEUE)
def collect_data(mo_id, report_form_id):
    u"""Асинхронная задача сбора данных по учреждению.

    :param mo_id: kinder.core.unit.models.Unit - Муниципальное образование.
    :param report_form_id: Отчет.
    :return: mo, data, exception
    """

    mo = Unit.objects.get(id=mo_id)
    report_form = ReportForm.objects.get(id=report_form_id)

    main_report_unit = report_form.unit
    unit_helper = UnitHelper(main_report_unit)

    data = None
    exception = None

    try:
        data = DataCollector(mo, unit_helper).collect(report_form)
    # Отложенное исключение
    except Exception as e:
        exception = e
    return mo.id, data, exception


@celery_app.task(queue=gs_settings.CELERY_WORKER_TASK_QUEUE)
def send_mo_data(mo_id, report_form_id, **kw):
    u"""Асинхронная задача отправки данных по учреждению.

    :param mo_id: kinder.core.unit.models.Unit - Муниципальное образование.
    :param report_form_id: Отчет.
    :param kw: Дополнительные параметры.
    :return: mo, exception
    """

    mo = Unit.objects.get(id=mo_id)
    report_form = ReportForm.objects.get(id=report_form_id)
    schedule_settings = ScheduleSettings.get_settings()
    exception = None
    try:
        send_report(report_form, schedule_settings, report_form.unit, mo)
    except (SOAPClientCreateException,
            ConnectingAtSendingReportException) as exc:
        # если произошла ошибка связанная с соединением
        # сервера сервиса, то перезапускаем таску через
        # "countdown" максимум "max_retries" раз
        try:
            send_mo_data.retry(
                exc=exc,
                max_retries=kw.get('max_retries'),
                countdown=kw.get('countdown'))
        except (MaxRetriesExceededError, ):
            # если все возможные попытки не увенчались успехом,
            # то сообщаем возвращаем исходную ошибку "exc"
            exception = exc
    except Exception as e:
        exception = e

    return mo.id, exception


def send_report(report_form, schedule_settings, main_report_unit, unit):
    u"""Функция выполяет отправку данных по учреждению.

    :param report_form: ReportForm отчет.
    :param schedule_settings: Настройки ФедОтч.
    :param main_report_unit: Главное учреждение в отчете.
    :param unit: Учреждение по которому отправляются данные.
    :return: list список не отправленных МО.
    """

    if not unit:
        unit = main_report_unit
    elif unit.kind.id not in [UnitKind.MO, UnitKind.REGION]:
        raise SendReportException()

    # Список МО по которым не удалось отправить данные
    not_send_mo = []
    for mo in unit.get_descendants(
            include_self=True).filter(kind__id=UnitKind.MO):

        unit_helper = UnitHelper(main_report_unit)
        try:
            document = TemplateView(
                mo, report_form.date,
                schedule_settings,
                unit_helper,
                report_form
            )
        except NoDataInMo as e:
            not_send_mo.append(mo.name)
            Logger.add_record(e.message, level=Logger.WARNING)
            continue
        except Exception as e:
            Logger.add_record(e.message, level=Logger.ERROR)
            raise DocumentCreateException()

        try:
            xml_content = render_xml_template(schedule_settings, document)
        except Exception as e:
            Logger.add_record(e.message, level=Logger.ERROR)
            raise XMLRenderException()

        Transport(schedule_settings).send(xml_content)

    return not_send_mo


def render_xml_template(schedule_settings, document):
    u"""Функция подготавливает данные к отправке.

    1) Рендерим шаблон документом document.
    2) Сжимаем, при включенной настройке.
    3) Кодируем в base64.
    """
    xml_string = loader.render_to_string(
        'ws/gisdo_request_template_4_0.xml', {'doc': document}).encode('utf-8')
    if schedule_settings.zip_xml:
        xml_string = zlib.compress(xml_string)[2:-4]

    encoded_xml = base64.b64encode(xml_string)

    return encoded_xml


class ReportService(ReportCacheMixin):
    def __init__(
            self, main_unit,
            set_progress_func=None,
            user=None):

        super(ReportService, self).__init__()

        self._unit_helper = UnitHelper(main_unit)
        self._report_form = ReportForm(
            user=user, unit=main_unit,
            in_progress=True)
        self._set_progress = set_progress_func or (lambda x: x)
        self._schedule_settings = ScheduleSettings.get_settings()

    def collect(self, send=True):
        self._report_form.save()

        # Счетчик кол-ва ошибок при отправке
        send_error_count = 0

        # На этом уровне перебираем учреждения типа МО
        current_mo_num = 1
        mo_count = len(list(self._unit_helper.get_report_units()))

        # Создаем группу параллельно выполняющихся подзадач
        # Каждая задача отвечает за подсчет показателей отдельного МО
        collect_data_tasks = group(
            collect_data.s(
                mo.id, self._report_form.id
            ) for mo in self._unit_helper.get_report_units())

        collect_data_tasks_results = collect_data_tasks(
            queue=gs_settings.CELERY_WORKER_TASK_QUEUE)

        mo_for_send = []

        # Обрабатываем результаты завершившихся задач
        for (mo_id, task_result, exception) in collect_data_tasks_results.iterate():
            self._set_progress(
                u'Обработано МО: %s из %s' % (
                    current_mo_num, mo_count))

            mo = Unit.objects.get(id=mo_id)

            now = str(datetime.datetime.now())

            try:
                # Выбрасываем отложенное исключение
                if exception:
                    raise exception
            except Exception as e:
                self._set_progress(
                    values={
                        now: (
                            u'Во время сбора данных по'
                            u' %s произошла ошибка: %s' % (
                                mo.name, e.message))
                    })
                current_mo_num += 1
                continue
            else:
                self._update_data(task_result)

            self._set_progress(
                values={
                    now: u'Данные по %s успешно собраны' % mo.name
                })

            current_mo_num += 1

            if send:
                mo_for_send.append(mo)

        current_mo_num = 1

        if send:
            self._set_progress(u'Подготовка к отправке данные МО')

            minutes, max_retries = (
                self._schedule_settings.get_interval_and_count_retries())

            # Создаем группу параллельно выполняющихся подзадач
            # Каждая задача отвечает за отправку показателей отдельного МО
            # Замечание: если подзадача выполнилась с ошибкой связанной с
            #   недоступностью сервера, то она будет перезапущена через
            #   "countdown" секунд максимум "max_retries" раз.
            #   И только по завершению всех возможных попыток всех подзадач
            #   группа подзадач считается выполненной.
            send_data_tasks = group(
                send_mo_data.s(
                    mo.id, self._report_form.id,
                    countdown=minutes * 60, max_retries=max_retries
                ) for mo in mo_for_send)

            # добавляем задержку каждой подзадачи на 1 секунду
            send_data_tasks_results = send_data_tasks.skew(
                start=1, stop=len(mo_for_send))(
                    queue=gs_settings.CELERY_WORKER_TASK_QUEUE)

            for (mo_id, exception) in send_data_tasks_results.iterate():
                self._set_progress(
                    u'Отправлены данные МО: %s из %s' % (
                        current_mo_num, mo_count))

                mo = Unit.objects.get(id=mo_id)
                now = str(datetime.datetime.now())

                try:
                    if exception:
                        raise exception
                except Exception as e:
                    self._set_progress(
                        values={
                            now: (
                                u'При передаче данных по %s '
                                u'произошла ошибка: %s' % (mo.name, e.message))
                        })
                    current_mo_num += 1
                    send_error_count += 1
                    continue

                self._set_progress(
                    values={
                        now: (
                            u'Данные по %s успешно отправлены' % mo.name)
                    })

                current_mo_num += 1

        if self._unit_helper.is_region:
            ReportFormRow.objects.get_or_create(
                report=self._report_form,
                unit=self._unit_helper.get_main_unit(),
                applications=json.dumps(self._cache[self.APPLICATIONS]),
                queues=json.dumps(
                    self._cache[self.QUEUES],
                    cls=AverageWaitingTimeWrapperEncoder),
                enrolled=json.dumps(self._cache[self.ENROLLED]),
                capacities=json.dumps(self._cache[self.CAPACITIES]))

        self._report_form.in_progress = False

        # Если все пакеты были отправлены то считаем что отчет отправлен
        if send and send_error_count == 0:
            self._report_form.sent_time = datetime.datetime.now()

        self._report_form.date = datetime.datetime.now()
        self._report_form.save()

    def _update_data(self, mo_data):
        if self._unit_helper.is_region:
            self._update_cache(mo_data[self.APPLICATIONS], self.APPLICATIONS)
            self._update_cache(mo_data[self.QUEUES], self.QUEUES)
            self._update_cache(mo_data[self.ENROLLED], self.ENROLLED)
            self._update_cache(mo_data[self.CAPACITIES], self.CAPACITIES)


class IndicesProcessor(object):
    """
    Генерирует правильный вид показателей для вставки их в xml
    """

    AGE_CATEGORIES_MAP = {
        const.ALL: 'all',

        const.AGE_CATEGORIES_FULL['2-6-MONTHS']: 'h_00_05',
        const.AGE_CATEGORIES_FULL['0.5-1-YEARS']: 'h_05_10',
        const.AGE_CATEGORIES_FULL['1-1.5-YEARS']: 'h_10_15',
        const.AGE_CATEGORIES_FULL['1.5-2-YEARS']: 'h_15_20',
        const.AGE_CATEGORIES_FULL['2-2.5-YEARS']: 'h_20_25',
        const.AGE_CATEGORIES_FULL['2.5-3-YEARS']: 'h_25_30',
        const.AGE_CATEGORIES_FULL['3-3.5-YEARS']: 'h_30_35',
        const.AGE_CATEGORIES_FULL['3.5-4-YEARS']: 'h_35_40',
        const.AGE_CATEGORIES_FULL['4-4.5-YEARS']: 'h_40_45',
        const.AGE_CATEGORIES_FULL['4.5-5-YEARS']: 'h_45_50',
        const.AGE_CATEGORIES_FULL['5-5.5-YEARS']: 'h_50_55',
        const.AGE_CATEGORIES_FULL['5.5-6-YEARS']: 'h_55_60',
        const.AGE_CATEGORIES_FULL['6-6.5-YEARS']: 'h_60_65',
        const.AGE_CATEGORIES_FULL['6.5-7-YEARS']: 'h_65_70',
        const.AGE_CATEGORIES_FULL['7-7.5-YEARS']: 'h_70_75',
        const.AGE_CATEGORIES_FULL['7.5-99-YEARS']: 'h_75_e',

        const.AGE_CATEGORIES_EIGHT['2-1-YEARS']: 'y_0_1',
        const.AGE_CATEGORIES_EIGHT['1-2-YEARS']: 'y_1_2',
        const.AGE_CATEGORIES_EIGHT['2-3-YEARS']: 'y_2_3',
        const.AGE_CATEGORIES_EIGHT['3-4-YEARS']: 'y_3_4',
        const.AGE_CATEGORIES_EIGHT['4-5-YEARS']: 'y_4_5',
        const.AGE_CATEGORIES_EIGHT['5-6-YEARS']: 'y_5_6',
        const.AGE_CATEGORIES_EIGHT['6-7-YEARS']: 'y_6_7',
        const.AGE_CATEGORIES_EIGHT['7-7.5-YEARS']: 'y_7_e',
    }

    @staticmethod
    def process(dirty_indices):

        result = []

        def get_valid_index(index):
            return "ind_{0}".format(index.replace('.', '_'))

        def get_valid_age_category(age_category):
            try:
                return IndicesProcessor.AGE_CATEGORIES_MAP[age_category]
            except KeyError:
                return age_category

        for dirty_index in dirty_indices:
            index = {
                'name': get_valid_index(dirty_index),
                'age_categories': [],
            }

            for dirty_age_category in dirty_indices[dirty_index]:
                index['age_categories'].append({
                    'name': get_valid_age_category(dirty_age_category),
                    'value': str(dirty_indices[dirty_index][dirty_age_category]),
                })

            result.append(index)

        return result


class CounterProxy(object):
    @staticmethod
    def transform(counter):

        result = []
        for index_id in counter:

            index = {
                'id': index_id,
                'indexes': []
            }

            for age_cat in counter[index_id]:
                index['indexes'].append({
                    'category': age_cat,
                    'count': counter[index_id][age_cat]
                })

            result.append(index)

        return result


website_pattern = re.compile("((https?|ftp)\:\/\/)?(\S+)")
phone_pattern = re.compile("([0-9]{10}(,[0-9]{10})*)")
email_pattern = re.compile("(\S+@\S+\.\S+)")


class TemplateView(object):
    u"""Объект, который рендерится в шаблон XML."""

    def __init__(self, mo, date, schedule_settings, unit_helper, report_form):

        self._mo = mo
        self._date = date.date()
        self._gisdo_settings = schedule_settings
        self._unit_helper = unit_helper
        self._report_form = report_form

    @property
    def header(self):
        u"""Данные для заголовка XML.

        То, что передается в теге system.
        """

        return {
            'system_name': self._gisdo_settings.system_name,
            'version': gs_settings.KINDER_VERSION,
            'email': self._gisdo_settings.email,
            'install_type': (
                self._gisdo_settings.installation_type
                if self._gisdo_settings.installation_type
                else InstallationType.REGIONAL
            ),
        }

    @property
    def municipality_unit(self):
        u"""Данные по муниципальным образованиям."""

        try:
            ext_unit = self._mo.ext_unit
        except EmieUnitModel.DoesNotExist:
            ext_unit = None

        (
            no_doo_act_0_3, no_doo_act_3_7,
            no_doo_def_0_3, no_doo_def_3_7,
            medic_0_3, medic_3_7,
            family_0_3, family_3_7
        ) = self.municipality_common_info()

        return {
            'oktmo': self._mo.octmo or u'нет',
            'epgu_link': self._mo.epgu_link or u'нет',
            'rpgu_link': self._mo.rpgu_link or u'нет',
            'name_mouo': (
                self._mo.full_name if self._mo.full_name else self._mo.name),
            'site_mouo': (
                self._mo.site
                if self._mo.site and
                website_pattern.match(self._mo.site) is not None
                else u'нет'),
            'address_mouo': self._mo.address_full or u'нет',
            'email_mouo': (
                self._mo.email
                if self._mo.email and
                email_pattern.match(self._mo.email) is not None
                else u'нет'),
            'phones_mouo': (
                self._mo.telephone
                if self._mo.telephone and
                phone_pattern.match(self._mo.telephone) is not None
                else u'нет'),
            'regulation': get_file_url(ext_unit, "service_info_doc"),
            'fix_area': get_file_url(ext_unit, "addr_binding_doc"),
            'no_doo_act_0_3': no_doo_act_0_3,
            'no_doo_act_3_7': no_doo_act_3_7,
            'no_doo_def_0_3': no_doo_def_0_3,
            'no_doo_def_3_7': no_doo_def_3_7,
            'medic_0_3': medic_0_3,
            'medic_3_7': medic_3_7,
            'family_0_3': family_0_3,
            'family_3_7': family_3_7
        }

    def municipality_common_info(self):
        u"""Данные, которые будут рендерятся в тег common."""

        today = datetime.date.today()

        no_doo_act_0_3 = Declaration.objects.filter(
            children__date_of_birth__gte=today - relativedelta(years=3),
            children__date_of_birth__lt=today,
            mo=self._mo,
            is_registered_in_absence=True
        ).filter(
            Q(desired_date__lte=DateMixin._get_current_calendar_year()) | Q(
                desired_date__isnull=True)
        ).exclude(
            status__code=DSS.ARCHIVE
        ).count()
        no_doo_act_3_7 = Declaration.objects.filter(
            children__date_of_birth__gte=today - relativedelta(years=7),
            children__date_of_birth__lte=today - relativedelta(years=3),
            mo=self._mo,
            is_registered_in_absence=True
        ).filter(
            Q(desired_date__lte=DateMixin._get_current_calendar_year()) | Q(
                desired_date__isnull=True)
        ).exclude(
            status__code=DSS.ARCHIVE
        ).count()

        no_doo_def_0_3 = Declaration.objects.filter(
            children__date_of_birth__gte=today - relativedelta(years=3),
            children__date_of_birth__lt=today,
            mo=self._mo,
            is_registered_in_absence=True
        ).filter(
            Q(desired_date__gt=DateMixin._get_current_calendar_year())
        ).exclude(
            status__code=DSS.ARCHIVE
        ).count()
        no_doo_def_3_7 = Declaration.objects.filter(
            children__date_of_birth__gte=today - relativedelta(years=7),
            children__date_of_birth__lte=today - relativedelta(years=3),
            mo=self._mo,
            is_registered_in_absence=True
        ).filter(
            Q(desired_date__gt=DateMixin._get_current_calendar_year())
        ).exclude(
            status__code=DSS.ARCHIVE
        ).count()

        medic_0_3 = Declaration.objects.filter(
            children__date_of_birth__gte=today - relativedelta(years=3),
            children__date_of_birth__lt=today,
            mo=self._mo,
            is_not_attending=True
        ).exclude(
            status__code=DSS.ARCHIVE
        ).count()
        medic_3_7 = Declaration.objects.filter(
            children__date_of_birth__gte=today - relativedelta(years=7),
            children__date_of_birth__lte=today - relativedelta(years=3),
            mo=self._mo,
            is_not_attending=True,
        ).exclude(
            status__code=DSS.ARCHIVE
        ).count()

        family_0_3 = Declaration.objects.filter(
            children__date_of_birth__gte=today - relativedelta(years=3),
            children__date_of_birth__lt=today,
            mo=self._mo,
            is_preschool_education=True
        ).exclude(
            status__code=DSS.ARCHIVE
        ).count()
        family_3_7 = Declaration.objects.filter(
            children__date_of_birth__gte=today - relativedelta(years=7),
            children__date_of_birth__lte=today - relativedelta(years=3),
            mo=self._mo,
            is_preschool_education=True
        ).exclude(
            status__code=DSS.ARCHIVE
        ).count()

        return (no_doo_act_0_3, no_doo_act_3_7,
                no_doo_def_0_3, no_doo_def_3_7,
                medic_0_3, medic_3_7,
                family_0_3, family_3_7)

    @property
    def units(self):
        u"""Данные по садам внутри муниципального образования."""

        # Признак наличия садов по которым можно
        # собрать данные
        dou_list = self._unit_helper.get_mo_units(self._mo.id)
        have_data = len(dou_list)

        if not have_data:
            raise NoDataInMo()

        for dou in dou_list:
            yield DouData(dou, self._report_form)


class DouData(object):
    u"""Класс формирует данные по ДОУ, которые рендерятся в шаблон."""

    def __init__(self, unit, report_form):
        self._unit = unit
        self._report_form = report_form

    @property
    def organization(self):
        u"""Данные для рендеринга в тег organization."""

        num_filial = get_filial_num(self._unit)

        if Group.objects.filter(
                unit=self._unit, status=GroupStatusEnum.FACT).exists():
            num_building = num_filial + 1
        else:
            num_building = num_filial

        return {
            'code': self._unit.gisdo.doo_identity,
            'name': self._unit.name,
            'type': get_dou_type(self._unit),
            'status': get_dou_status(self._unit),
            'structure': get_dou_structure_type(self._unit),
            'director': self._unit.boss_fio or u'нет',
            'worktime': (
                self._unit.reception_time or u"с 6.30 до 18.30 часов"),
            'license': 1 if self._unit.have_lic else 2,
            'type_area': (
                2 if self._unit.locality == LocalityEnumerate.RURAL else 1),
            'fias_org_guid': get_unit_fias_house_guid(self._unit),
            'org_address': (
                self._unit.address_full
                if self._unit.address_full
                else u'нет'),
            'website': (
                self._unit.site
                if self._unit.site and
                website_pattern.match(self._unit.site) is not None
                else u'нет'),
            'email': (
                self._unit.email
                if self._unit.email and
                email_pattern.match(self._unit.email) is not None
                else u'нет'),
            'phone': (
                self._unit.telephone
                if self._unit.telephone and
                phone_pattern.match(self._unit.telephone) is not None
                else u'нет'),
            'additional_education': get_additional_info_by_dou(self._unit),
            'features': get_additional_info_by_dou(
                self._unit, model=DopService),
            'num_filial': num_filial,
            'num_building': num_building,
            'num_group': get_groups_num(self._unit),
            'partner_doo': 2 if self._unit.have_lic else 1
        }

    @property
    def indices(self):
        u"""Показатели."""
        rows = ReportFormRow.objects.filter(
            unit=self._unit, report=self._report_form)

        indices = []
        for row in rows:
            index = {}
            index.update(json.loads(row.applications))
            index.update(json.loads(row.queues))
            index.update(json.loads(row.enrolled))
            index.update(json.loads(row.capacities))
            indices.extend(IndicesProcessor.process(index))

        return indices

    @property
    def buildings(self):
        """Данные по зданиям."""

        if Group.objects.filter(
                unit=self._unit, status=GroupStatusEnum.FACT).exists():

            type_building = (
                self._unit.building_type
                if self._unit.building_type else BuildingTypes.TYPICAL)
            filial_attr = (
                0 if type_building == 3 else get_filial_info(self._unit))

            yield {
                'fias_house_guid': get_unit_fias_house_guid(self._unit),
                'id': '1' + str(self._unit.id),
                'name': self._unit.name,
                'plain_address': (
                    self._unit.address_full
                    if self._unit.address_full
                    else u'нет'),
                'depreciation': (
                    self._unit.percent_of_wear
                    if self._unit.percent_of_wear
                    else '15'),
                'building_type_area': (
                    2 if self._unit.locality == LocalityEnumerate.RURAL else 1),
                'type_building': type_building,
                'filial': filial_attr,
                'groups': GroupData(self._unit).groups
            }

        filials = FilialData.objects.filter(
            head=self._unit
        ).filter(
            UnitHelper.ALLOWED_FILIAL_FILTER
        ).select_related("filial")

        for filial_data in filials:
            filial = filial_data.filial
            if Group.objects.filter(
                    unit=filial, status=GroupStatusEnum.FACT).exists():
                type_building = (
                    filial.building_type
                    if filial.building_type else BuildingTypes.TYPICAL)
                filial_attr = 0 if type_building == 3 else 1

                yield {
                    'fias_house_guid': get_unit_fias_house_guid(filial),
                    'id': filial.id,
                    'name': filial.name,
                    'plain_address': (
                        filial.address_full
                        if filial.address_full
                        else u'нет'),
                    'depreciation': (
                        filial.percent_of_wear
                        if filial.percent_of_wear
                        else '15'),
                    'building_type_area': (
                        2 if self._unit.locality == LocalityEnumerate.RURAL else 1),
                    'type_building': type_building,
                    'filial': filial_attr,
                    'groups': GroupData(filial).groups
                }


class GroupData(object):
    u"""Данные по группам для рендеринга в шаблон."""

    OVZ_VALUES = {
        HNE.DEAFNESS: 1,
        HNE.HARDOFHEARTING: 1,
        HNE.SPEACH: 2,
        HNE.PHONETICS: 2,
        HNE.BLINDNESS: 3,
        HNE.AMAUROSIS: 3,
        HNE.BACKLIGHT: 4,
        HNE.BACKHARD: 4,
        HNE.BACK: 5,
        HNE.DISABLEMENT: 6,
        HNE.INVALIDITY: 7,
        'other': 8
    }

    WELLNESS = {
        HNE.PHTHISIS: 1,
        HNE.SICK: 2,
        'other': 3
    }

    WORKTYPE = {
        WorkType.SHORT: 1,
        WorkType.ABREV: 2,
        WorkType.FULL: 3,
        WorkType.EXTEND: 4,
        WorkType.ALLDAY: 5,
    }

    def __init__(self, dou):
        self._dou = dou
        self._group_index = Index(dou)

    @staticmethod
    def get_age_range(group):
        u"""Возрастная категория группы."""
        def get_range(x):
            u"""Разбиваем отрезок на age_from, age_to."""
            return map(int, x.split('-'))

        if group.sub_age_cat:
            age_from, age_to = get_range(group.sub_age_cat.code)
        elif group.age_cat:
            age_from, age_to = get_range(group.age_cat.code)
        else:
            age_from, age_to = 1, 7

        if not age_from:
            age_from += 0.1

        return age_from, age_to

    @classmethod
    def _get_partner_info(cls, group):
        u"""Фукнция формирующая информацию о партнерстве группы."""

        dou = group.unit

        if dou.have_lic or group.edu_unit is None:
            return 2, u"нет"
        else:
            return 1, group.edu_unit

    @classmethod
    def _get_group_ovz_type(cls, group):
        if group.health_need:
            return cls.OVZ_VALUES.get(
                group.health_need.code, cls.OVZ_VALUES['other'])
        else:
            return cls.OVZ_VALUES['other']

    @classmethod
    def _get_group_wellness(cls, group):
        if group.health_need:
            return cls.WELLNESS.get(
                group.health_need.code, cls.WELLNESS['other'])
        else:
            return cls.WELLNESS['other']

    @classmethod
    def _get_group_orientation(cls, group):
        """Возвращает значение для тэга orientation(Направленность группы)"""
        orientation_value = {
            GroupTypeEnumerate.DEV: {'orientation': 1},
            GroupTypeEnumerate.COMP: {
                'orientation': 2,
                'ovzType': cls._get_group_ovz_type(group)
            },
            GroupTypeEnumerate.COMBI: {
                'orientation': 3,
                'ovzType': cls._get_group_ovz_type(group)
            },
            GroupTypeEnumerate.HEALTH: {
                'orientation': 4,
                'wellness': cls._get_group_wellness(group)
            },
            'is_young_children': {"orientation": 5},
            'is_care': {"orientation": 6},
            'is_family': {"orientation": 7},
        }

        if (group.type and (
                    group.type.code in (
                        GroupTypeEnumerate.COMP,
                        GroupTypeEnumerate.COMBI,
                        GroupTypeEnumerate.HEALTH))):
            return orientation_value[group.type.code]

        if group.is_family:
            return orientation_value['is_family']

        if group.is_care:
            return orientation_value['is_care']

        if group.is_young_children:
            return orientation_value['is_young_children']

        return orientation_value[GroupTypeEnumerate.DEV]

    @classmethod
    def _get_group_work_time(cls, group):
        """Возвращает значение режима работы группы для фед. отчет."""

        def normalize_work_type_code(code):
            if code in [WorkType.ABREV, WorkType.ALLDAY,
                        WorkType.EXTEND, WorkType.FULL, WorkType.SHORT]:
                return code
            else:
                try:
                    torn_code = code.split("-")
                except AttributeError:
                    # Код может быть не заполнен
                    return WorkType.FULL

            map_code = torn_code[0]
            map_code = map_code.replace(",", ".")

            try:
                numeric_map_code = float(map_code)
            except ValueError:
                # Если код не число, то передаем "Полного дня (10,5-12 часов)"
                return WorkType.FULL

            if 1 <= numeric_map_code < 8:
                code_compliance = WorkType.SHORT
            elif 8 <= numeric_map_code < 10.5:
                code_compliance = WorkType.ABREV
            elif 10.5 <= numeric_map_code < 12.5:
                code_compliance = WorkType.FULL
            elif 12.5 <= numeric_map_code < 24:
                code_compliance = WorkType.EXTEND
            elif numeric_map_code == 24:
                code_compliance = WorkType.ALLDAY
            else:
                code_compliance = WorkType.FULL

            return code_compliance

        if group.work_type:
            normalize_code = normalize_work_type_code(group.work_type.code)
            return cls.WORKTYPE[normalize_code]
        else:
            return cls.WORKTYPE[WorkType.FULL]

    @property
    def groups(self):
        u"""Данные по группам."""
        for group in Group.objects.filter(
                status=GroupStatusEnum.FACT, unit=self._dou).order_by('id'):

            age_from, age_to = self.get_age_range(group)
            data = {
                'id': group.id,
                'ageFrom': age_from,
                'ageTo': age_to,
                'ovzType': '',
                'wellness': '',
                'add_cont': len(self._group_index.directs_in_group(group.id)),
                'capacity': self._group_index.get_capacity_in_group(
                    force=False, group_id=group.id),
                'subgroup': (
                    group.subgroup_number if group.subgroup_number else 1),
                'transfer_space': group.vacant_places,
                'enrolled': len(self._group_index.enrolled_in_group(
                    force=False, group_id=group.id, query_filter="default")),
                'work_time_group': self._get_group_work_time(group),
                'free_space': self._group_index.free_space_in_group(
                    force=False, group=group)
            }

            data.update(self._get_group_orientation(group))
            # Количество детей, у которых поле "ОВЗ" заполнено,
            # все кроме пусто и значения "Нет"
            if data["orientation"] in (1, 5, 6, 7):
                data["ovz_deti"] = len(
                    self._group_index.enrolled_in_group(
                        force=False, group_id=group.id,
                        query_filter="with_hn"
                    )
                )
            elif data["orientation"] in (2, 4):
                data["ovz_deti"] = data["enrolled"]
            elif data["orientation"] == 3:
                data["ovz_deti"] = len(
                    self._group_index.enrolled_in_group(
                        force=False, group_id=group.id,
                        query_filter="by_disabled_1"
                    )
                )

            if data["orientation"] in [5, 6]:
                data["activity"] = 2
            else:
                data["activity"] = 1

            if group.name:
                data.update({'name': group.name})
            else:
                data.update({"name": u"нет"})

            partner_group, partner = self._get_partner_info(group)
            data["partner_group"] = partner_group
            data["partner"] = partner

            yield data


class Transport(object):
    def __init__(self, schedule_settings):

        self._settings = schedule_settings
        self._client = self._create_client()

    def _create_client(self):

        try:
            header = Element('Authentication', ns=('ns1', 'http://eo.edu.ru'))
            header.append(
                Element('Login', ns=('ns1', 'http://eo.edu.ru')).setText(
                    self._settings.push_login))
            header.append(
                Element('Password', ns=('ns1', 'http://eo.edu.ru')).setText(
                    self._settings.push_password))
            client = Client(
                url=gs_settings.GISDO_WS_WSDL,
                timeout=gs_settings.GISDO_TIMEOUT)
            client.set_options(soapheaders=(header,))
        except Exception:
            raise SOAPClientCreateException()
        return client

    def _get_send_method(self):

        if self._settings.zip_xml:
            if self._settings.async_send:
                method = 'PushZipDataAsync'
            else:
                method = 'PushZipData'
        else:
            method = 'PushData'

        return method

    def _get_number_of_attempts(self):

        return int(self._settings.resend_count)

    def send(self, content):

        method = self._get_send_method()

        attempts = 1
        error_msg = None
        error_inst = None
        while attempts <= self._get_number_of_attempts():
            try:
                result = getattr(self._client.service, method)(
                    data=content, schema=u'3.0')
            except (socket.timeout, suds.WebFault) as error:
                error_inst = error
                error_msg = str(error).encode('utf-8')
                attempts += 1
            except Exception, error:
                error_inst = error
                error_msg = str(error).encode('utf-8')
                break
            else:
                if not result['Result']:
                    error_msg = result['Message']
                break

        if error_msg:
            Logger.add_record(error_msg, level=Logger.ERROR)
            if isinstance(error_inst, (socket.timeout, suds.WebFault)):
                raise ConnectingAtSendingReportException()
            raise SendReportException()
